;;;
;;;   common lisp package
;;;
;;;         Mikio Yamamoto (1990).
;;;



;;;  ----------
;;;  SETF MACRO
;;;  ----------

(macro (setf exp)
  (let ((left (macroexpand (cadr exp)))
        (right (caddr exp)))
    (cond ((symbol? left) `(set! ,left ,right))
          ((pair? left)
           (let ((update-fn (getprop (car left) 'setf-macro)))
             (cond ((symbol? update-fn)
                    `(,(getprop (car left) 'setf-macro)
                       ,@(cdr left)
                       ,right))
                   ((procedure? update-fn)
                    (apply update-fn (append (cdr left) (list right))))
                   (else (error "Illegal argument.~%        In ~s."
                                (list 'setf left right))))))
          (else
            (error "Illegal argument.~%        In ~s."
                   (list 'setf left right))))))

(macro (defsetf exp)
  (let ((acc (cadr exp))
	(sub-or-vars (caddr exp))
	(body (cdddr exp)))
    (cond ((symbol? sub-or-vars)
	   (if (null? body)
	       `(begin (putprop ',acc 'setf-macro ',sub-or-vars) ',acc)
	       (error "defsetf: too many argument.(type1 defsetf)")))
	  ((pair? sub-or-vars)
	   `(begin
	      (putprop ',acc 'setf-macro
		       (lambda ,(append sub-or-vars (car body))
			 ,@(cdr body)))
	      ',acc))
	  (else
	   (error "defsetf: 2nd argument is irregular.")))))


(defsetf string-ref string-set!)
(defsetf vector-ref vector-set!)
(defsetf list-ref (l n) (x) `(set-car! (list-tail ,l ,n) ,x))
(defsetf list-tail (l n) (x) `(set-car! (list-tail ,l ,n) ,x))

(macro (set-c???r! exp)
  (let* ((name (symbol->string (car exp)))
	 (left (cadr exp))
	 (right (caddr exp)))
    (case (string-ref name 5)
	  ((#\a) `(set-car! (,(make-place name) ,left) ,right))
	  ((#\d) `(set-cdr! (,(make-place name) ,left) ,right)))))

(define (make-place name)
  (do ((pos 6 (1+ pos))
       (ad-list nil))
      ((eq? (string-ref name pos) #\r)
       (string->symbol (apply string `(#\c ,@(reverse ad-list) #\r))))
    (set! ad-list (cons (string-ref name pos) ad-list))))

(define set-caar! set-c???r!)
(define set-cadr! set-c???r!)
(define set-cdar! set-c???r!)
(define set-cddr! set-c???r!)

(define set-caaar! set-c???r!)
(define set-caadr! set-c???r!)
(define set-cadar! set-c???r!)
(define set-caddr! set-c???r!)
(define set-cdaar! set-c???r!)
(define set-cdadr! set-c???r!)
(define set-cddar! set-c???r!)
(define set-cdddr! set-c???r!)

(define set-caaaar! set-c???r!)
(define set-caaadr! set-c???r!)
(define set-caadar! set-c???r!)
(define set-caaddr! set-c???r!)
(define set-cadaar! set-c???r!)
(define set-cadadr! set-c???r!)
(define set-caddar! set-c???r!)
(define set-cadddr! set-c???r!)
(define set-cdaaar! set-c???r!)
(define set-cdaadr! set-c???r!)
(define set-cdadar! set-c???r!)
(define set-cdaddr! set-c???r!)
(define set-cddaar! set-c???r!)
(define set-cddadr! set-c???r!)
(define set-cdddar! set-c???r!)
(define set-cddddr! set-c???r!)

(defsetf car set-car!)
(defsetf cdr set-cdr!)
(defsetf caar set-caar!)
(defsetf cadr set-cadr!)
(defsetf cdar set-cdar!)
(defsetf cddr set-cddr!)

(defsetf caaar set-caaar!)
(defsetf caadr set-caadr!)
(defsetf cadar set-cadar!)
(defsetf caddr set-caddr!)
(defsetf cdaar set-cdaar!)
(defsetf cdadr set-cdadr!)
(defsetf cddar set-cddar!)
(defsetf cdddr set-cdddr!)

(defsetf caaaar set-caaaar!)
(defsetf caaadr set-caaadr!)
(defsetf caadar set-caadar!)
(defsetf caaddr set-caaddr!)
(defsetf cadaar set-cadaar!)
(defsetf cadadr set-cadadr!)
(defsetf caddar set-caddar!)
(defsetf cadddr set-cadddr!)
(defsetf cdaaar set-cdaaar!)
(defsetf cdaadr set-cdaadr!)
(defsetf cdadar set-cdadar!)
(defsetf cdaddr set-cdaddr!)
(defsetf cddaar set-cddaar!)
(defsetf cddadr set-cddadr!)
(defsetf cdddar set-cdddar!)
(defsetf cddddr set-cddddr!)


;;;
;;;   -----------
;;;   list as set
;;;   -----------
;;;

(define (subset? small-set large-set)
  (every? (lambda (element)
	    (memq element large-set))
	  small-set))

(define (duplicate? set)
  (let ((subset set))
    (some? (lambda (element)
	     (set! subset (cdr subset))
	     (memq element subset))
	   set)))



;;;
;;; ---------
;;; DEFSTRUCT
;;; ---------
;;;

(macro (defstruct args)
  `(begin
     (defstruct-type ',(cadr args) ,(length (cddr args)))
     (eval ',(make-define-make-fun (cadr args) (cddr args)))
     (eval '(define (,(string->symbol
                        (string-append (symbol->string (cadr args)) "?")) x)
              (and (structure? x)
                   (eq? (structure-type x) ',(cadr args)))))
     ,@(do ((slot (cddr args) (cdr slot))
	    (ith 0 (1+ ith))
	    (result ()))
	   ((null? slot) result)
	 (set! result 
	       (cons (make-define-access-fun (cadr args) (car slot) ith)
		     result))
	 (set! result
	       (cons (make-define-setfmacro (cadr args) (car slot) ith)
		     result)))
     ',(cadr args)))


;;; (make-make-fun-sym struct-name)
(macro (make-make-fun-sym exp)
  `(string->symbol (string-append "make-" (symbol->string ,(cadr exp)))))

;;; (make-key-slot-sym slot-name)
(macro (make-key-slot-sym exp)
  `(string->symbol (string-append ":" (symbol->string ,(cadr exp)))))

;;; (make-access-fun-sym struct-name slot-name)
(macro (make-access-fun-sym exp)
  `(string->symbol (string-append (symbol->string ,(cadr exp))
				  "-"
				  (symbol->string ,(caddr exp)))))

(define (make-define-access-fun struct-name slot ith)
  (if (pair? slot) (set! slot (car slot)))
  `(eval '(define (,(make-access-fun-sym struct-name slot) structure)
	    (structure-ref ',struct-name structure ,ith))))

(define (make-define-setfmacro struct-name slot ith)
  `(defsetf
     ,(make-access-fun-sym struct-name
			   (if (pair? slot) (car slot) slot))
     (structure) (right)
     (list 'structure-set! '',struct-name structure ,ith right)))

(define (make-define-make-fun struct-name elements)
  `(macro (,(make-make-fun-sym struct-name) exp)
     (cons 'structure (cons '',struct-name
			    (get-structure-val ',(keylization elements)
					       (cdr exp))))))

(define (keylization elements)
  (map (lambda (slot)
	 (if (pair? slot)
	     (cons (make-key-slot-sym (car slot)) (cdr slot))
	     (make-key-slot-sym slot)))
       elements))


(define (get-structure-val slots args)
  (letrec ((slot-list (eventh-list args))
	   (val-list (oddth-list args))
	   (len (length args))
	   (val-decision 
	    (lambda (slot-data)
	      (let ((slot-name (if (pair? slot-data)(car slot-data) slot-data))
		    (slot-default (if (pair? slot-data) (cadr slot-data) ())))
		(if (memq slot-name slot-list)
		    (some? (lambda (arg-slot val)
			     (if (eq? arg-slot slot-name) val))
			   slot-list
			   val-list)
		    slot-default)))))
    (cond ((odd? len) 
	   (error "make-structure: incomplete value spec(the number of args is odd)."))
	  ((not (subset? slot-list
			 (map (lambda (a) (if (pair? a) (car a) a)) slots)))
	   (error "make-structure: unknown slot-name(s)."))
	  ((duplicate? slot-list)
	   (error "make-structure: duplicate slot-name."))
	  (else
	   (map (lambda (slot) (val-decision slot)) slots)))))
		

;;;

(define (eventh-list list)
  (do ((rest list (cddr rest))
       (result ()))
      ((null? rest) (reverse result))
    (set! result (cons (car rest) result))))

(define (oddth-list list)
  (eventh-list (cdr list)))
