;;
;;    macro
;;

(macro (dlet form)
  (let ((temps (map (lambda (def) (gensym)) (cadr form)))
         (returnval (gensym)) )
    `(let ,(map (lambda (temp def) `(,temp ,(car def)))
		temps (cadr form))
	  ,@(map (lambda (def) `(set! ,@def))
		 (cadr form))
       (let ((,returnval (begin ,@(cddr form))))
	  ,@(map (lambda (temp def) `(set! ,(car def) ,temp))
		 temps (cadr form))
	  ,returnval))))

(macro (dolist form)
  (let ((sym (gensym)))
    `(do ((,sym ,(cadr (cadr form)) (cdr ,sym)))
         ((null? ,sym) ,@(cddr (cadr form)))
       (let ((,(caadr form) (car ,sym)) )
	  ,@(cddr form)) ))) 

(macro (dotimes form)
  (let ((sym (gensym)))
    `(let ((,sym ,(cadr (cadr form))))
       (do ((,(caadr form) 0 (1+ ,(caadr form))))
           ((>= ,(caadr form) ,sym) ,@(cddr (cadr form)))
         ,@(cddr form)))))
	      
(macro (let/cc form)
  `(call/cc (lambda (,(cadr form)) ,@(cddr form))))

(macro (time form)
  `(let ((result '())
	 (start-time 0) (end-time 0)
	 (real-start 0) (real-end 0)
	 (utime 0) (stime 0))
     (set! real-start (get-time))
     (set! start-time (ptime))
     (set! result ,(cadr form))
     (set! end-time (ptime))
     (set! real-end (get-time))
     (set! utime (time-diff (car start-time) (car end-time)))
     (set! stime (time-diff (cdr start-time) (cdr end-time)))
     (format #t "total time: ~a secs~%" (time-diff2 real-start real-end))
     (format #t "user time: ~a secs~%" (usec-to-msec2 utime))
     (format #t "system time: ~a secs~%" (usec-to-msec2 stime))
     result))

(macro (autoload form)
  (if (not (bound? `,(cadr (cadr form))))
      (eval `(define ,(cadr (cadr form))
	       (lambda exp
		 (load ,(caddr form))
		 (apply ,(cadr (cadr form)) exp)))))
  `,(cadr form))
