(define *defined-funs* '())
(define *redirected-funs* '())
(define *rd-vars* '())

(define (redirect forms)
  (dlet ((*defined-funs* '())
         (*redirected-funs* '()))
    (for-each rp1 forms)
    (let ((redirected-forms (map rp2 forms)))
      (append (map rp3 *redirected-funs*)
              redirected-forms))))

(define (redirect-file name)
  (let* ((name-string (if (symbol? name) (symbol->string name) name))
         (in-name (string-append name-string ".lsp"))
         (out-name (string-append name-string ".out")))
    (set! *defined-funs* '())
    (set! *redirected-funs* '())
    (set! *rd-vars* '())
    (format t "Redirecting from ~a to ~a.~&" in-name out-name)
    (call-with-input-file in-name
      (lambda (in)
        (do ((form (read in) (read in)))
            ((eof-object? form))
          (write-raw-char #\*) (flush-port (current-output-port))
          (rp1 form))))
    (format t "End pass 1.~&")
    (let ((in (open-input-file in-name))
          (out (open-output-file out-name)))
      (do ((form (read in) (read in)))
          ((eof-object? form))
        (write-raw-char #\*) (flush-port (current-output-port))
        (let* ((prev-redirected-funs *redirected-funs*)
               (redirected-form (rp2 form)))
          (do ((x *redirected-funs* (cdr x)))
              ((eq? x prev-redirected-funs))
            (format out "~s~&" (rp3 (car x))))
          (format out "~s~&" redirected-form)))
      (close-input-port in)
      (close-output-port out))))

(define (rp1 form)
  (if (and (pair? form) (eq? (car form) 'define))
      (set! *defined-funs*
            (cons (if (symbol? (cadr form)) (cadr form) (caadr form))
                  *defined-funs*))))

(define (rp2 form)
  (rp2expr form))

(define (rp3 pair)
  `(define ,(cdr pair) (copy-function ,(car pair))))

(define (rp2expr form)
  (if (pair? form)
      (let ((fun (car form)) (args (cdr form)))
        (if (symbol? fun)
            (case fun
              ((if and or begin delay) (cons fun (rp2args args)))
              ((lambda) (rp2lambda args))
              ((set!) (list fun (car args) (rp2expr (cadr args))))
              ((let) (rp2let args))
              ((let*) (rp2let* args))
              ((letrec) (rp2letrec args))
              ((do) (rp2do args))
              ((dolist dotimes) (rp2dolist fun args))
              ((define)
               (if (pair? (car args))
                   (list* 'define
                          (car args)
                          (cddr (rp2lambda (cons (cdar args) (cdr args)))))
                   (list* 'define (car args) (rp2args (cdr args)))))
              ((quote) form)
              ((cond) (cons fun (map rp2args args)))
              ((case) 
               (list* fun 
                      (rp2expr (car args))
                      (map (lambda (clause)
                             (cons (car clause) (rp2args (cdr clause))))
                           (cdr args))))
              ((quasiquote) (rp2quasiquote args))
              ((dlet) (rp2dlet args))
              ((macro)
               (format t "Found a macro definition: ~s~&" form))
              (else (if (macro-function fun)
                        (begin
                          (format t "Macro form: ~s~&" form)
                          (rp2expr ((macro-function fun) form)))
                        (cons (rp2redirect fun)
                              (rp2args args)))))
            (rp2args form)))
      form))

(define (rp2redirect name)
  (if (or (member name *rd-vars*)
          (member name *defined-funs*))
      name
      (let ((x (assoc name *redirected-funs*)))
        (if x
            (cdr x)
            (let ((y (string->symbol
                       (string-append "$$" (symbol->string name)))))
              (set! *redirected-funs* 
                (cons (cons name y) *redirected-funs*))
              y)))))

(define (rp2bind name)
  (set! *rd-vars* (cons name *rd-vars*)))

(define (rp2args forms)
  (if (end? forms)
      '()
      (cons (rp2expr (car forms)) (rp2args (cdr forms)))))

(define (rp2body body)
  (dlet ((*rd-vars* *rd-vars*))
    (do ((x body (cdr x)))
        ((or (null? x)
             (not (pair? (car x)))
             (not (eq? (caar x) 'define))))
      (rp2bind (if (pair? (cadar x)) (caadar x) (cadar x))))
    (rp2args body)))

(define (rp2lambda args)
  (define (rp2lambda-bind x)
    (cond ((symbol? x) (rp2bind x))
          ((null? x))
          (else (rp2bind (car x))
                (rp2lambda-bind (cdr x)))))
  (dlet ((*rd-vars* *rd-vars*))
    (rp2lambda-bind (car args))
    (list* 'lambda (car args) (rp2body (cdr args)))))

(define (rp2let args)
  (dlet ((*rd-vars* *rd-vars*))
    (if (symbol? (car args))
        (begin
          (rp2bind (car args))
          (let ((bind (map (lambda (x) 
                             (list (car x) (rp2expr (cadr x))))
                           (cadr args))))
            (dolist (x (cadr args)) (rp2bind (car x)))
            (list* 'let (car args) bind (rp2body (cddr args)))))
        (let ((bind (map (lambda (x) (list (car x) (rp2expr (cadr x))))
                         (car args))))
          (dolist (x (car args)) (rp2bind (car x)))
          (list* 'let bind (rp2body (cdr args)))))))

(define (rp2let* args)
  (dlet ((*rd-vars* *rd-vars*))
    (list* 'let*
           (map (lambda (x)
                  (rp2bind (car x))
                  (list (car x) (rp2expr (cadr x))))
                (car args))
           (rp2body (cdr args)))))

(define (rp2letrec args)
  (dlet ((*rd-vars* *rd-vars*))
    (dolist (x (car args)) (rp2bind (car x)))
    (list* 'letrec
           (map (lambda (x) (list (car x) (rp2expr (cadr x))))
                (car args))
           (rp2body (cdr args)))))

(define (rp2do args)
  (dlet ((*rd-vars* *rd-vars*))
    (let ((bind
            (map (lambda (x)
                   (list* (car x) (rp2expr (cadr x)) (cddr x)))
                 (car args))))
      (dolist (x (car args)) (rp2bind (car x)))
      (list* 'do
        (map (lambda (x)
               (if (null? (cddar args))
                   x
                   (list (car x) (cadr x) (rp2expr (caddr x)))))
             bind)
        (rp2args (cadr args))
        (rp2body (cddr args))))))

(define (rp2dolist name args)
  (dlet ((*rd-vars* *rd-vars*))
    (let ((x (rp2expr (cadar args))))
      (rp2bind (caar args))
      (list* name
        (if (null? (cddar args))
            (list (caar args) x)
            (list (caar args) x (rp2expr (caddar args))))
        (rp2body (cdr args))))))

(define (rp2quasiquote args)
  (define (qq x level)
    (cond ((vector? x)
           (list->vector (qq (vector->list x) level)))
          ((not (pair? x)) x)
          ((member (car x) '(unquote unquote-splicing))
           (list (car x)
                 (if (zero? level)
                     (rp2expr (cadr x))
                     (qq (cadr x) (1- level)))))
          ((eq? (car x) 'quasiquote)
           (list (car x) (qq (cadr x) (1+ level))))
          (else
            (cons (qq (car x) level) (qq (cdr x) level)))))
  (list 'quasiquote (qq (car args) 0)))

(define (rp2dlet args)
  (list* 'dlet
         (map (lambda (x) (list (car x) (rp2expr (cadr x))))
              (car args))
         (rp2body (cdr args))))
