;;;
;;;     scheme.lsp
;;;

(set! *max-back-trace-depth* 10000000000000000)

(set! *default-kanji-input-code* 'euc)
(set! *default-kanji-output-code* 'euc)
(format #t "The value of *inline-functions* is ~a.~%"
        (set! *inline-functions* #f))
(format #t "The value of *print-circle* is ~a.~%"
        (set! *print-circle* #f))

(if (bound? 'thread-id)
    (set! *gc-message-level* 1)
    (set! *debug-level* 2))

;;; PROVIDE and REQUIRE

(define provide '())
(define require '())

(let ((*modules* '()))
  (set! provide  (lambda (module)
		   (if (not (member module *modules*))
		       (set! *modules* (cons module *modules*)))))
  (set! require (lambda (module . filenames)
		  (if (not (member module *modules*))
		      (if (null? filenames)
			  (load module)
			  (for-each load filenames))))))

;;; compile function for system library

(define (make-syslib-file file)
  (if (not (symbol? file))
      (error "Argument must be symbol."))
  (let ((opc (print-circle #f)))
    (cond ((memq file '(sclib topfun debugfun compiler mod-compiler
                        mcomp-t mcomp-nt mcomp-ltc-nt re-compiler
                        mcomp-she-t
                        compiler_pvm topfun_pvm))
           (let ((x *inline-functions*))
             (set! *inline-functions* #t)
             (load 'redirect)
             (redirect-file file)
             (system
               (string-append "mv "
                 (string-append (symbol->string file) ".out ")
                 (string-append (symbol->string file) ".out.lsp")))
             (compile-file (string-append (symbol->string file) ".out"))
             (system
               (string-append "mv "
                 (string-append (symbol->string file) ".out.fas ")
                 (string-append (symbol->string file) ".fas")))
             (set! *inline-functions* x)))
          ((memq file '(maclib common process help err_system))
           (let ((x *inline-functions*))
             (set! *inline-functions* #t)
             (compile-file file)
             (set! *inline-functions* x))))
    (print-circle opc)))

;;; PWD, CD, PUSHD, POPD, DIRS and LS

(define (expand-file-name2 filename)
  (expand-file-name
    (if (string=? filename "~")
        "~/"
        filename)))

(define (pwd)
  (or (getwd) "pwd: can't stat ."))

(define cd '())
(define pushd '())
(define popd '())
(define dirs '())

(let ((*dir-stack* (list (pwd))))
  (set! cd
    (lambda arg
      (and (if (null? arg)
               (chdir (getenv "HOME"))
               (chdir (expand-file-name2 (car arg))))
           (set-car! *dir-stack* (pwd))
           (pwd))))
  (set! pushd
    (lambda arg
      (if (null? arg) 
          (let ((tmp (and (pair? (cdr *dir-stack*)) (cadr *dir-stack*))))
            (if (not tmp)
                (format #t "pushd: No other directory.")
                (begin
                  (set-car! (cdr *dir-stack*) (car *dir-stack*))
                  (set-car! *dir-stack* tmp)
                  (chdir (expand-file-name2 (car *dir-stack*)))
                  *dir-stack*)))
          (let ((x (chdir (expand-file-name2 (car arg)))))
            (if x
                (set! *dir-stack* (cons (pwd) *dir-stack*))
                #f)))))
  (set! popd
    (lambda ()
      (if (null? (cdr *dir-stack*))
          (format #t "popd: Directory stack empty.")
          (begin
            (set! *dir-stack* (cdr *dir-stack*))
            (chdir (expand-file-name2 (car *dir-stack*)))
            *dir-stack*))))
  (set! dirs (lambda () *dir-stack*)))

(define (ls . arg)
  (system
    (string-append "ls -F " (if (null? arg) "" (car arg)))))

;;; misc

(define (com exp)
  (write-code (compile exp) (current-output-port)))

(define random
  (let ((seed 1)
	(a (+ (* 3141 1000 1000) (* 592 1000) 621))
	(m (expt 2 63))
	(c 1))
    (lambda (k)
      (if (not (integer? k))
	  (error "RANDOM is valid only for integers %a" k))
      (if (not (and (positive? k) (<= k m)))
	  (error "RANDOM is valid only for integers from 1 to %a" m))
      (set! seed (remainder (+ (* a seed) c) m))
      (quotient (* seed k) m))))

(define (tuslook key)
  (system (format #f "tuslook ~a" key)))

(define (make-list n . init)
  (let ((init (if (pair? init) (car init) '())))
    (define (aux n x)
      (if (zero? n)
          x
          (aux (1- n) (cons init x))))
    (aux n '())))

;;; TRACE-safe call/cc

(define (ts-call/cc f)
  (call/cc
    (lambda (k)
      (f (let ((trace-level *trace-level*))
           (lambda (v)
             (set! *trace-level* trace-level)
             (k v)))))))

;;;

(autoload 'fact '~komiya/scheme/parallel/fact)
(autoload 'fib '~komiya/scheme/parallel/fib)
(autoload 'tak '~komiya/scheme/parallel/tak)
(autoload 'finger '~komiya/scheme/misc/finger)

(define (slib)
  (load "/pro/tus/scheme-app/slib/tus.init"))

;(require 'debug)
;(macro (trace form) `(define ,(cadr form) (tracef (eval ,(cadr form)))))
;(macro (untrace form) `(define ,(cadr form) (untracef (eval ,(cadr form)))))

(autoload 'ncomp "~komiya/scheme/misc/native-comp.lsp")

;;; for PVMTUS

(if (bound? 'spawntus)
    (set! *re-compile-level* 3))

;;; tus [-f<filename>]*

(and (bound? '*argv*)
     (dolist (x (cdr *argv*))
       (and (> (string-length x) 2)
            (char=? (string-ref x 0) #\-)
            (char=? (string-ref x 1) #\f)
            (let ((fn (list->string (cddr (string->list x)))))
              (if (file-is-readable? fn)
                  (load fn)
                  (format #t "Warning: Cannot open the file ~a.~%" fn))))))

(autoload 'cab-1 '~/scheme/cltc/cpab8)

;;; another read-eval-print-loop
(and (bound? 'readline)
     (not (bound? 'spawntus))
;     (not (bound? 'exectcl))
     (load (string-append (getenv "SCM_DIR") "/readexp.lsp")))  ; never return
