;;;(macro (defmacro form)
;;;  `(let ((fun (lambda ,@(cddr form)))
;;;	 (macro ,(cadr form)
;;;	    (lambda (par) (apply fun (cdr par)))) )))

(define (protect-symbol sym)
  (set! *protect-symbols* (cons sym *protect-symbols*)))
  
(define aaa (protect-symbol 'dlet))
(define aaa (protect-symbol 'dolist))
(define aaa (protect-symbol 'dotimes))
(define aaa (protect-symbol 'gensym))
(define aaa (protect-symbol 'let/cc))

(define (gensym)
  (let ((num -1))
      (lambda ()
         (set! num (1+ num))
         (str->sym-nointern (string-append "#$" (number->string num))))))

(define gensym (gensym))


;;;;;;;;;;
;;;
;;;    user error break function
;;;

(define aaa (protect-symbol 'error))

(define (error s . args)
  (format (console-output-port) "~&")
  (apply format (console-output-port) s args)
  (error-break))



;;;;;;;;;;
;;;
;;;    loader
;;;

(define aaa (protect-symbol '*loading-message*))
(define aaa (protect-symbol 'load))

(define *loading-message* #t)

(define (load filename . switch)
  (set! switch (if (or (null? switch) (not (car switch))) #f #t))
  (let ((full-filename (make-full-filename filename)))
	 ;;; full-filename = (filename . file-kind)
	 ;;;                             file-kind = lsp or fas or ()
    (if (not (_file-exists? (car full-filename)))
        (error 'file-error       ;;; By N.Watanabe
               'function load
               'arguments (list filename switch)
               'pathname (car full-filename)
               'message (format #f "Cannot open the file ~a." 
				   (car full-filename)))
        (if (eq? (cdr full-filename) 'fas)
          (begin
            (if *loading-message*
                (format #t "Loading ~a...~%" (car full-filename)))
            (bcload (car full-filename))
            (if *loading-message*
                (begin (display "Finished.") (newline)))
            (car full-filename))
          (begin
            (if *loading-message*
              (format #t "Loading ~a...~%" (car full-filename)))
            (_call-with-input-file (car full-filename)
              (lambda (p)
                (do ((form (read p) (read p)))
                  ((eof-object? form)
                   (if *loading-message*
                     (begin (display "Finished.") (newline)))
                   (car full-filename))
                  (if switch
                    (begin (display (eval form)) (newline))
                    (eval form))))))))))

;(define (fas-eval form)
;  (if (and (pair? form)(pair? (cdr form)) (pair? (cddr form))
;	   (pair? (caddr form))
;	   (eq? (car (caddr form)) 'compiled-function))
;      (bctran-define-form form)
;      (eval form)))

(define (make-full-filename filename)
  (let* ((str-name (cond ((symbol? filename) (symbol->string filename))
			 ((string? filename) filename)
			 (else (error 'simple-error
				      'format-string "~s is not a string or a symbol."
				      'format-arguments (list filename)
				      'function make-full-filename
				      'arguments (list filename))))))
    (set! str-name (expand-file-name str-name))
    (case (get-file-extender str-name)
      ((lsp) (cons str-name 'lsp))
      ((fas) (cons str-name 'fas))
      (else  (cond ((or (string? filename) (_file-exists? str-name))
		    (cons str-name '()))
		   ((_file-exists? (string-append str-name ".fas"))
		    (cons (string-append str-name ".fas") 'fas))
      		   ((_file-exists? (string-append str-name ".lsp"))
		    (cons (string-append str-name ".lsp") 'lsp))
		   (else (cons str-name '())))))))

(define (make-filename filename)
  (let* ((str-name (cond ((symbol? filename) (symbol->string filename))
			 ((string? filename) filename)
			 (else (error 'simple-error
				      'format-string "~s is not a string or a symbol."
				      'format-arguments (list filename)
				      'function make-filename
				      'arguments (list filename))))))
    (expand-file-name str-name)))

(define (expand-file-name)
  (define (take-path n str)
    (let* ((i (- (string-length str) n))
           (new-str (make-string i)))
      (do ((j 0 (1+ j)))
          ((>= j i) new-str)
        (string-set! new-str j (string-ref str (+ j n))))))
  (define (gethdir str)
    (do ((i 1 (1+ i)) (len (string-length str)))
        ((or (>= i len)
             (char=? (string-ref str i) #\/))
         (let ((s (make-string (1- i))))
           (do ((j 0 (1+ j)))
               ((>= (1+ j) i)
                (let ((t (make-string (- len i))))
                  (do ((k 0 (1+ k))
                       (j (1+ j) (1+ j)))
                      ((>= j len) (cons s t))
                    (string-set! t k (string-ref str j)))))
             (string-set! s j (string-ref str (1+ j))))))))
  (lambda (fname)
    (if (char=? (string-ref fname 0) #\~)
        (if (or (= (string-length fname) 1)
                (char=? (string-ref fname 1) #\/))
            (string-append (getenv "HOME") (take-path 1 fname))
            (let* ((d (gethdir fname))
                   (p (getpwnam (car d))))
              (if (not p)
                  (error "Unknown user: ~a~%" (car d))
                  (string-append p (cdr d)))))
        fname)))

(define expand-file-name (expand-file-name))

(define (get-file-extender str)
  (let ((strlength (string-length str)))
    (cond ((< strlength 4) nil )
	  ((and (eqv? #\. (string-ref str (- strlength 4)))
		(eqv? #\l (string-ref str (- strlength 3)))
		(eqv? #\s (string-ref str (- strlength 2)))
		(eqv? #\p (string-ref str (- strlength 1))))
	   'lsp)
	  ((and (eqv? #\. (string-ref str (- strlength 4)))
		(eqv? #\f (string-ref str (- strlength 3)))
		(eqv? #\a (string-ref str (- strlength 2)))
		(eqv? #\s (string-ref str (- strlength 1))))
	   'fas)
	  (else nil))))

(define aaa (protect-symbol 'call-with-input-file))
(define aaa (protect-symbol 'call-with-output-file))
(define aaa (protect-symbol 'open-input-file))
(define aaa (protect-symbol 'open-output-file))
(define aaa (protect-symbol 'file-exists?))
(define aaa (protect-symbol 'file-is-directory?))
(define aaa (protect-symbol 'file-is-readable?))
(define aaa (protect-symbol 'file-is-writable?))
(define aaa (protect-symbol 'file-is-executable?))
(define aaa (protect-symbol 'transcript-on))
(define aaa (protect-symbol 'transcript-off))

(define (call-with-input-file fname func)
  (_call-with-input-file (make-filename fname) func))

(define (call-with-output-file fname func)
  (_call-with-output-file (make-filename fname) func))

(define (open-input-file fname)
  (_open-input-file (make-filename fname)))

(define (open-output-file fname)
  (_open-output-file (make-filename fname)))

(define (file-exists? fname)
  (_file-exists? (make-filename fname)))

(define (file-is-directory? fname)
  (_file-is-directory? (make-filename fname)))

(define (file-is-readable? fname)
  (_file-is-readable? (make-filename fname)))

(define (file-is-writable? fname)
  (_file-is-writable? (make-filename fname)))

(define (file-is-executable? fname)
  (_file-is-executable? (make-filename fname)))

(define transcript-on '())
(define transcript-off '())

(let ((filename '()))
  (set! transcript-on
	(lambda (fname)
	  (let ((f (make-filename fname)))
	    (_transcript-on f)
	    (set! filename f)
	    filename)))
  (set! transcript-off
	(lambda ()
	  (if (_transcript-off)
	      filename
	      #f))))


;;;
;;;    editor
;;;

(define aaa (protect-symbol 'ed))

(define (ed . fname)
  (let ((load-file (if (null? fname) (mine)
		       (mine (make-mine-filename (car fname))))))
    (if load-file (load load-file))))

(define (make-mine-filename filename)
  (let* ((str-name (cond ((symbol? filename) (symbol->string filename))
			 ((string? filename) filename)
			 (else (error 'simple-error
				      'format-string "Argument must be a string or a symbol."
				      'format-arguments '()
				      'function make-mine-filename
				      'arguments (list filename))))))
    (set! str-name (expand-file-name str-name))
    (case (get-file-extender str-name)
      ((lsp fas) str-name)
      (else
        (if (or (string? filename) (_file-exists? str-name))
	    str-name
	    (string-append str-name ".lsp"))))))


;;;;;;;;;; 
;;; 
;;;    mesuring function execution time 
;;;
;;;        (exec-time function . argument)
;;;

(define eval '())

(define aaa (protect-symbol 'exec-time))

(define (exec-time func . arg)
  (let ((result 0) (start-time 0) (end-time 0) (utime 0) (stime 0))
    (set! start-time (ptime))
    (set! result  (apply func arg))
    (set! end-time (ptime))
    (set! utime (time-diff (car start-time) (car end-time)))
    (set! stime (time-diff (cdr start-time) (cdr end-time)))
    (list 'result: result
	  'total-time: (usec-to-msec (time-add utime stime))
	  'user-time: (usec-to-msec utime)
	  'system-time: (usec-to-msec stime))))

(define (time-diff stime etime)
  (let ((sec (- (car etime) (car stime)))
	(msec (- (cdr etime) (cdr stime))))
    (cond ((< msec 0)
	   (cons (1- sec) (+ msec 1000000)))
	  (else
	   (cons sec msec)))))

(define (time-add time1 time2)
  (let ((sec (+ (car time1) (car time2)))
	(msec (+ (cdr time1) (cdr time2))))
    (cond ((>= msec 1000000)
	   (cons (1+ sec) (- msec 1000000)))
	  (else
	   (cons sec msec)))))

(define (usec-to-msec time)
  (cons (car time) (/ (cdr time) 1000)))

;;;
;;;    (time <exp>)
;;;

(define (usec-to-msec2 time)
  (+ (car time) (/ (cdr time) 1000000.0)))

(define (time-diff2 start-time end-time)
  (- (+ (* (vector-ref end-time 0) 60 60.0)
	(* (vector-ref end-time 1) 60)
	(vector-ref end-time 2)
	(/ (vector-ref end-time 3) 100))
     (+ (* (vector-ref start-time 0) 60 60.0)
	(* (vector-ref start-time 1) 60)
	(vector-ref start-time 2)
	(/ (vector-ref start-time 3) 100))))

;;;
;;; format
;;;
;;; Common Lisp Interface on Ultimate Scheme. Takafumi Kawana. 89/06/14 ;;;
;;; 89/11/11 format, dolist, dotimes					;;;
;;; 90/06/11 format removed		Akira Matsushita		;;;

(define aaa (protect-symbol 'format))

(define (format flag format-string . args)
  (define (write-radix n r)
    (display (number->string n r)))
  (define (main)
    (do ((c (read-char) (read-char)))
        ((eof-object? c)
         (if (not flag) (get-output-string (current-output-port))))
      (if (char=? c #\~)
          (let ((c (read-char)))
            (case c
              ((#\~) (display c))
              ((#\%) (newline))
              ((#\&) (if (not (first-column? (current-output-port)))
			 (newline)))
              ((#\c) (write-char (car args)) (set! args (cdr args)))
              ((#\a) (display (car args)) (set! args (cdr args)))
              ((#\b) (write-radix (car args) 2) (set! args (cdr args)))
              ((#\o) (write-radix (car args) 8) (set! args (cdr args)))
              ((#\d) (write-radix (car args) 10) (set! args (cdr args)))
              ((#\x) (write-radix (car args) 16) (set! args (cdr args)))
              ((#\s) (write (car args)) (set! args (cdr args)))
              (else (write-char #\~) (write-char c))))
          (write-char c))))
  (with-output-to-port
    (case flag
      ((#f) (open-output-string))
      ((#t) (current-output-port))
      (else flag))
    (lambda () (with-input-from-port (open-input-string format-string) main))))

;;;
;;;  simple print function
;;;

(define (print2 . args)
  (do ((x args (cdr x)))   ((null? x))
     (if (string? (car x))
         (display (car x))
         (write (car x)))))

;;;
;;; misc
;;;

(define aaa (protect-symbol 'list->vector))
;(define aaa (protect-symbol 'list-tail))
;(define aaa (protect-symbol 'list-ref))
(define aaa (protect-symbol 'list->string))
(define aaa (protect-symbol 'string->list))
(define aaa (protect-symbol 'expt))
;(define aaa (protect-symbol 'nth))
;(define aaa (protect-symbol 'nthcdr))
;(define aaa (protect-symbol 'last))
(define aaa (protect-symbol 'nconc))
(define aaa (protect-symbol 'nreverse))
(define aaa (protect-symbol 'print-circle))

(define (list->vector x) (apply vector x))

;(define (list-tail x k)
;  (if (zero? k)
;      x
;      (list-tail (cdr x) (1- k))))

;(define (list-ref x k) (car (list-tail x k)))

;(define (nth n l) (list-ref l n))

;(define (nthcdr n l) (list-tail l n))

;(define (last x)
;  (letrec ((aux (lambda (l)
;                  (if (pair? (cdr l))
;                      (aux (cdr l))
;                      l))))
;    (if (pair? x) (aux x) x)))

(define nconc
  (lambda args
    (cond ((null? args) '())
          ((null? (cdr args)) (car args))
          ((null? (car args)) (apply nconc (cdr args)))
          (else
            (set-cdr! (last (car args))
                      (apply nconc (cdr args)))
            (car args)))))

(define (nreverse x)
  (cond ((null? x) x)
        ((not (list? x))
         (error "nreverse: Not a list in ~s~%" x))
        (else (do ((reved '() x)
                   (rev-cdr (cdr x) (cdr rev-cdr))
                   (x x rev-cdr))
                ((begin (set-cdr! x reved) (null? rev-cdr)) x)))))

(define (list->string l)
  (apply string l))

(define (string->list s)
  (do ((i (1- (string-length s)) (1- i))
       (l '() (cons (string-ref s i) l)))
      ((< i 0) l)))

(define (expt b n)
  (define (square n) (* n n))
  (define (int_expt b n)
    (cond ((zero? n) 1)
	  ((even? n) (square (int_expt b (/ n 2))))
	  (else (* b (int_expt b (1- n))))))
  (cond ((and (integer? n)
              (or (integer? b) (integer? (imag-part b))))
         (if (>= n 0)
             (int_expt b (truncate n))
             (/ 1 (int_expt b (truncate (- n))))))
        (else (_expt b n))))

(define (print-circle x)
  (let ((prev *print-circle*))
    (set! *print-circle* x)
    prev))
